"""
===============
Pipeline Schema
===============

This module is responsible for managing the "pipeline schemas", i.e. the allowable 
and fully supported pipelines.

"""

from collections.abc import Iterable
from pathlib import Path

from layered_config_tree import LayeredConfigTree

from easylink.graph_components import EdgeParams
from easylink.pipeline_schema_constants import ALLOWED_SCHEMA_PARAMS
from easylink.step import HierarchicalStep, NonLeafConfigurationState, Step


class PipelineSchema(HierarchicalStep):
    """All possible pipelines that are fully supported.

    A ``PipelineSchema`` is a :class:`~easylink.step.HierarchicalStep` whose
    :class:`~easylink.graph_components.StepGraph` determines all possible allowable pipelines.
    The fundamental purpose of this class is to validate that the user-requested
    pipeline to run conforms to a fully supported pipeline.

    See :class:`~easylink.step.HierarchicalStep` for inherited attributes.

    Parameters
    ----------
    name
        The name of the pipeline schema.
    nodes
        The nodes of the pipeline schema.
    edges
        The edges of the pipeline schema.

    Notes
    -----
    All ``PipelineSchema`` instances are intended to be created by the :meth:`_get_schemas`
    class method.

    The ``PipelineSchema`` is a high-level abstraction; it represents the desired
    pipeline of conceptual steps to run with no detail as to how each of those
    steps is implemented.

    """

    def __init__(self, name: str, nodes: Iterable[Step], edges: Iterable[EdgeParams]) -> None:
        super().__init__(name, nodes=nodes, edges=edges)

    def __repr__(self) -> str:
        return f"PipelineSchema.{self.name}"

    def validate_step(
        self, pipeline_config: LayeredConfigTree, input_data_config: LayeredConfigTree
    ) -> dict[str, list[str]]:
        """Validates the pipeline configuration against this ``PipelineSchema``.

        Parameters
        ----------
        pipeline_config
            The pipeline configuration to validate.
        input_data_config
            The input data configuration.

        Returns
        -------
            A dictionary of errors, where the keys are the names of any steps that
            did not validate and the values are lists of as many error messages as
            could be generated for each of those steps.

        Notes
        -----
        Below, we nest the full pipeline configuration under a "substeps" key of
        a root :class:`~easylink.step.HierarchicalStep` because such a root step
        doesn't exist from the user's perspective and doesn't appear explicitly in
        the user-provided pipeline specification file.
        """
        return super().validate_step(
            step_config={"substeps": pipeline_config["steps"]},
            combined_implementations=pipeline_config["combined_implementations"],
            input_data_config=input_data_config,
        )

    def validate_inputs(self, input_data: dict[str, Path]) -> dict[str, list[str]]:
        """Validates the file's existence and properties for each file slot.

        Parameters
        ----------
        input_data
            A dictionary mapping input data slot names to file paths.

        Returns
        -------
            A dictionary of errors, where the keys are the names of any files that
            did not validate and the values are lists of as many error messages as
            could be generated for each of those files.
        """
        errors = {}
        for _, _, edge_attrs in self.step_graph.out_edges("input_data", data=True):
            validator = edge_attrs["input_slot"].validator
            for file in input_data.values():
                try:
                    validator(file)
                except FileNotFoundError as e:
                    errors[str(file)] = ["File not found."]
                except Exception as e:
                    errors[str(file)] = [e.args[0]]
        return errors

    def configure_pipeline(
        self, pipeline_config: LayeredConfigTree, input_data_config: LayeredConfigTree
    ) -> None:
        """Configures the ``PipelineSchema`` and corresponding `StepGraphs<easylink.graph_components.StepGraph`.

        The configuration state of any :class:`~easylink.step.Step` tells whether
        that ``Step`` is a leaf or a non-leaf node and is assigned to the
        :attr:`easylink.step.Step.configuration_state`. By definition, the entire
        ``PipelineSchema`` has non-leaf configuration state; this method thus assigns
        a :class:`~easylink.step.NonLeafConfigurationState` to the ``PipelineSchema``.
        Upon instantiation, this ``NonLeafConfigurationState`` recursively updates
        the ``StepGraphs`` until all non-leaf nodes are resolved.

        Parameters
        ----------
        pipeline_config
            The pipeline configuration.
        input_data_config
            The input data configuration.
        """
        self._configuration_state = NonLeafConfigurationState(
            self,
            pipeline_config["steps"],
            combined_implementations=pipeline_config["combined_implementations"],
            input_data_config=input_data_config,
        )

    @classmethod
    def _get_schemas(cls) -> list["PipelineSchema"]:
        """Gets all allowable ``PipelineSchemas``.

        These ``PipelineSchemas`` represent the fully supported pipelines and are
        used to validate the user-requested pipeline.

        Returns
        -------
            All allowable ``PipelineSchemas``.
        """
        return [
            cls(name, nodes=nodes, edges=edges)
            for name, (nodes, edges) in ALLOWED_SCHEMA_PARAMS.items()
        ]


PIPELINE_SCHEMAS = PipelineSchema._get_schemas()
"""All allowable :class:`PipelineSchemas<PipelineSchema>` to validate the requested
pipeline against."""



import os
import re
import subprocess
import tempfile

import pytest
from pytest_mock import MockerFixture

from easylink.pipeline_schema import PipelineSchema
from easylink.pipeline_schema_constants import TESTING_SCHEMA_PARAMS
from easylink.runner import main
from easylink.utilities.general_utils import is_on_slurm
from tests.conftest import RESULTS_DIR, SPECIFICATIONS_DIR


@pytest.mark.slow
@pytest.mark.skipif(
    not is_on_slurm(),
    reason="Must be on slurm to run this test.",
)
def test_slurm(mocker: MockerFixture, caplog: pytest.LogCaptureFixture) -> None:
    """Test that the pipeline runs on SLURM with appropriate resources."""
    nodes, edges = TESTING_SCHEMA_PARAMS["integration"]
    mocker.patch("easylink.pipeline_schema.ALLOWED_SCHEMA_PARAMS", TESTING_SCHEMA_PARAMS)
    mocker.patch(
        "easylink.configuration.Config._get_schema",
        return_value=PipelineSchema("integration", nodes=nodes, edges=edges),
    )
    results_dir = tempfile.mkdtemp(dir=RESULTS_DIR)
    # give the tmpdir the same permissions as the parent directory so that
    # cluster jobs can write to it
    os.chmod(results_dir, os.stat(RESULTS_DIR).st_mode)
    with pytest.raises(SystemExit) as exit:
        main(
            command="run",
            pipeline_specification=SPECIFICATIONS_DIR / "integration/pipeline.yaml",
            input_data=SPECIFICATIONS_DIR / "common/input_data.yaml",
            computing_environment=SPECIFICATIONS_DIR
            / "integration/environment_spark_slurm.yaml",
            results_dir=results_dir,
            debug=True,
        )
    assert exit.value.code == 0
    output = caplog.text
    job_ids = re.findall(r"Job \d+ has been submitted with SLURM jobid (\d+)", output)
    assert len(job_ids) == 1
    cmd = [
        "sacct",
        "--jobs=" + ",".join(job_ids),
        "--format=JobID,Account,Partition,ReqMem,ReqCPUS,TimelimitRaw",
        "--noheader",
        "--parsable2",
    ]
    result = subprocess.run(cmd, capture_output=True, text=True)
    output = result.stdout
    # Filter out jobs that are not the main job and grab full line
    main_job_pattern = r"^(\d+)\|"
    main_line = None
    for line in output.split("\n"):
        if re.match(main_job_pattern, line):
            main_line = line
            break
    assert main_line
    fields = main_line.split("|")
    account, partition, mem, cpus, time = fields[1:6]
    assert account == "proj_simscience"
    assert partition == "all.q"
    assert mem == "1G" or mem == "1024M"  # Just in case
    assert cpus == "1"
    assert time == "60"



# mypy: ignore-errors
import os
import tempfile

import pytest

from easylink.pipeline_schema import PipelineSchema
from easylink.pipeline_schema_constants import TESTING_SCHEMA_PARAMS
from easylink.runner import main
from tests.conftest import RESULTS_DIR, SPECIFICATIONS_DIR


@pytest.mark.slow
def test_missing_results(mocker, caplog):
    """Test that the pipeline fails when a step is missing output files."""
    nodes, edges = TESTING_SCHEMA_PARAMS["integration"]
    mocker.patch("easylink.pipeline_schema.ALLOWED_SCHEMA_PARAMS", TESTING_SCHEMA_PARAMS)
    mocker.patch(
        "easylink.configuration.Config._get_schema",
        return_value=PipelineSchema("integration", nodes=nodes, edges=edges),
    )

    ## Mock implementation script call to wait 1s instead of running something
    mocker.patch(
        "easylink.implementation.Implementation.script_cmd",
        new_callable=mocker.PropertyMock,
        return_value="sleep 1s",
    )
    results_dir = tempfile.mkdtemp(dir=RESULTS_DIR)
    # give the tmpdir the same permissions as the parent directory so that
    # cluster jobs can write to it
    os.chmod(results_dir, os.stat(RESULTS_DIR).st_mode)
    with pytest.raises(SystemExit) as exit:
        main(
            command="run",
            pipeline_specification=SPECIFICATIONS_DIR / "integration" / "pipeline.yaml",
            input_data=SPECIFICATIONS_DIR / "common/input_data.yaml",
            computing_environment=SPECIFICATIONS_DIR / "common/environment_local.yaml",
            results_dir=results_dir,
        )
    assert exit.value.code == 1
    assert "MissingOutputException" in caplog.text



"""
=====================
Test Pipeline Schemas
=====================
"""

from easylink.graph_components import (
    EdgeParams,
    InputSlot,
    InputSlotMapping,
    OutputSlot,
    OutputSlotMapping,
)
from easylink.step import (
    HierarchicalStep,
    InputStep,
    LoopStep,
    OutputStep,
    ParallelStep,
    Step,
)
from easylink.utilities.validation_utils import validate_input_file_dummy

SINGLE_STEP_NODES = [
    InputStep(),
    Step(
        step_name="step_1",
        input_slots=[
            InputSlot(
                name="step_1_main_input",
                env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                validator=validate_input_file_dummy,
            )
        ],
        output_slots=[OutputSlot("step_1_main_output")],
    ),
    OutputStep(
        input_slots=[
            InputSlot(name="result", env_var=None, validator=validate_input_file_dummy)
        ],
    ),
]
SINGLE_STEP_EDGES = [
    EdgeParams(
        source_node="input_data",
        target_node="step_1",
        output_slot="all",
        input_slot="step_1_main_input",
    ),
    EdgeParams(
        source_node="step_1",
        target_node="results",
        output_slot="step_1_main_output",
        input_slot="result",
    ),
]

SINGLE_STEP_SCHEMA_PARAMS = (SINGLE_STEP_NODES, SINGLE_STEP_EDGES)

TRIPLE_STEP_NODES = [
    InputStep(),
    Step(
        step_name="step_1",
        input_slots=[
            InputSlot(
                name="step_1_main_input",
                env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                validator=validate_input_file_dummy,
            )
        ],
        output_slots=[OutputSlot("step_1_main_output")],
    ),
    Step(
        step_name="step_2",
        input_slots=[
            InputSlot(
                name="step_2_main_input",
                env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                validator=validate_input_file_dummy,
            )
        ],
        output_slots=[OutputSlot("step_2_main_output")],
    ),
    Step(
        step_name="step_3",
        input_slots=[
            InputSlot(
                name="step_3_main_input",
                env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                validator=validate_input_file_dummy,
            )
        ],
        output_slots=[OutputSlot("step_3_main_output")],
    ),
    OutputStep(
        input_slots=[
            InputSlot(name="result", env_var=None, validator=validate_input_file_dummy)
        ],
    ),
]
TRIPLE_STEP_EDGES = [
    EdgeParams(
        source_node="input_data",
        target_node="step_1",
        output_slot="all",
        input_slot="step_1_main_input",
    ),
    EdgeParams(
        source_node="step_1",
        target_node="step_2",
        output_slot="step_1_main_output",
        input_slot="step_2_main_input",
    ),
    EdgeParams(
        source_node="step_2",
        target_node="step_3",
        output_slot="step_2_main_output",
        input_slot="step_3_main_input",
    ),
    EdgeParams(
        source_node="step_3",
        target_node="results",
        output_slot="step_3_main_output",
        input_slot="result",
    ),
]

TRIPLE_STEP_SCHEMA_PARAMS = (TRIPLE_STEP_NODES, TRIPLE_STEP_EDGES)


BAD_COMBINED_TOPOLOGY_NODES = [
    InputStep(),
    LoopStep(
        template_step=HierarchicalStep(
            step_name="step_1",
            input_slots=[
                InputSlot(
                    name="step_1_main_input",
                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validator=validate_input_file_dummy,
                ),
            ],
            output_slots=[OutputSlot("step_1_main_output")],
            nodes=[
                Step(
                    step_name="step_1a",
                    input_slots=[
                        InputSlot(
                            name="step_1a_main_input",
                            env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                            validator=validate_input_file_dummy,
                        ),
                    ],
                    output_slots=[OutputSlot("step_1a_main_output")],
                ),
                Step(
                    step_name="step_1b",
                    input_slots=[
                        InputSlot(
                            name="step_1b_main_input",
                            env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                            validator=validate_input_file_dummy,
                        ),
                    ],
                    output_slots=[OutputSlot("step_1b_main_output")],
                ),
            ],
            edges=[
                EdgeParams(
                    source_node="step_1a",
                    target_node="step_1b",
                    output_slot="step_1a_main_output",
                    input_slot="step_1b_main_input",
                ),
            ],
            input_slot_mappings=[
                InputSlotMapping(
                    parent_slot="step_1_main_input",
                    child_node="step_1a",
                    child_slot="step_1a_main_input",
                ),
            ],
            output_slot_mappings=[
                OutputSlotMapping(
                    parent_slot="step_1_main_output",
                    child_node="step_1b",
                    child_slot="step_1b_main_output",
                ),
            ],
        ),
        self_edges=[
            EdgeParams(
                source_node="step_1",
                target_node="step_1",
                output_slot="step_1_main_output",
                input_slot="step_1_main_input",
            ),
        ],
    ),
    OutputStep(
        input_slots=[
            InputSlot(name="result", env_var=None, validator=validate_input_file_dummy)
        ],
    ),
]

BAD_COMBINED_TOPOLOGY_SCHEMA_PARAMS = (BAD_COMBINED_TOPOLOGY_NODES, SINGLE_STEP_EDGES)


NESTED_TEMPLATED_STEPS_NODES = [
    InputStep(),
    LoopStep(
        template_step=ParallelStep(
            template_step=HierarchicalStep(
                step_name="step_1",
                input_slots=[
                    InputSlot(
                        name="step_1_main_input",
                        env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                        validator=validate_input_file_dummy,
                    ),
                ],
                output_slots=[OutputSlot("step_1_main_output")],
                nodes=[
                    Step(
                        step_name="step_1a",
                        input_slots=[
                            InputSlot(
                                name="step_1a_main_input",
                                env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                                validator=validate_input_file_dummy,
                            ),
                        ],
                        output_slots=[OutputSlot("step_1a_main_output")],
                    ),
                    Step(
                        step_name="step_1b",
                        input_slots=[
                            InputSlot(
                                name="step_1b_main_input",
                                env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                                validator=validate_input_file_dummy,
                            ),
                        ],
                        output_slots=[OutputSlot("step_1b_main_output")],
                    ),
                ],
                edges=[
                    EdgeParams(
                        source_node="step_1a",
                        target_node="step_1b",
                        output_slot="step_1a_main_output",
                        input_slot="step_1b_main_input",
                    ),
                ],
                input_slot_mappings=[
                    InputSlotMapping(
                        parent_slot="step_1_main_input",
                        child_node="step_1a",
                        child_slot="step_1a_main_input",
                    ),
                ],
                output_slot_mappings=[
                    OutputSlotMapping(
                        parent_slot="step_1_main_output",
                        child_node="step_1b",
                        child_slot="step_1b_main_output",
                    ),
                ],
            ),
        ),
        self_edges=[
            EdgeParams(
                source_node="step_1",
                target_node="step_1",
                output_slot="step_1_main_output",
                input_slot="step_1_main_input",
            ),
        ],
    ),
    OutputStep(
        input_slots=[
            InputSlot(name="result", env_var=None, validator=validate_input_file_dummy)
        ],
    ),
]


NESTED_TEMPLATED_STEPS_SCHEMA_PARAMS = (NESTED_TEMPLATED_STEPS_NODES, SINGLE_STEP_EDGES)


COMBINE_WITH_ITERATION_NODES = [
    InputStep(),
    LoopStep(
        template_step=Step(
            step_name="step_1",
            input_slots=[
                InputSlot(
                    name="step_1_main_input",
                    env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                    validator=validate_input_file_dummy,
                )
            ],
            output_slots=[OutputSlot("step_1_main_output")],
        ),
        self_edges=[
            EdgeParams(
                source_node="step_1",
                target_node="step_1",
                output_slot="step_1_main_output",
                input_slot="step_1_main_input",
            ),
        ],
    ),
    Step(
        step_name="step_2",
        input_slots=[
            InputSlot(
                name="step_2_main_input",
                env_var="DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                validator=validate_input_file_dummy,
            )
        ],
        output_slots=[OutputSlot("step_2_main_output")],
    ),
    OutputStep(
        input_slots=[
            InputSlot(name="result", env_var=None, validator=validate_input_file_dummy)
        ],
    ),
]
DOUBLE_STEP_EDGES = [
    EdgeParams(
        source_node="input_data",
        target_node="step_1",
        output_slot="all",
        input_slot="step_1_main_input",
    ),
    EdgeParams(
        source_node="step_1",
        target_node="step_2",
        output_slot="step_1_main_output",
        input_slot="step_2_main_input",
    ),
    EdgeParams(
        source_node="step_2",
        target_node="results",
        output_slot="step_2_main_output",
        input_slot="result",
    ),
]


COMBINE_WITH_ITERATION_SCHEMA_PARAMS = (COMBINE_WITH_ITERATION_NODES, DOUBLE_STEP_EDGES)



"""
========
Pipeline
========

This module is responsible for the ``Pipeline`` class, whose primary purpose
is to perform validations as well as generate the Snakefile to be used by 
`Snakemake <https://snakemake.readthedocs.io/en/stable/>`_ to execute the pipeline.

"""

from collections import defaultdict
from pathlib import Path

from loguru import logger

from easylink.configuration import Config
from easylink.pipeline_graph import PipelineGraph
from easylink.rule import (
    AggregationRule,
    CheckpointRule,
    ImplementedRule,
    InputValidationRule,
    TargetRule,
)
from easylink.utilities.general_utils import exit_with_validation_error
from easylink.utilities.paths import SPARK_SNAKEFILE
from easylink.utilities.validation_utils import validate_input_file_dummy

IMPLEMENTATION_ERRORS_KEY = "IMPLEMENTATION ERRORS"


class Pipeline:
    """A convenience class for validations and Snakefile generation.

    Parameters
    ----------
    config
        The :class:`~easylink.configuration.Config` object.

    Attributes
    ----------
    config
        The :class:`~easylink.configuration.Config` object.
    pipeline_graph
        The :class:`~easylink.pipeline_graph.PipelineGraph` object.
    spark_is_required
        A boolean indicating whether the pipeline requires Spark.
    any_embarrassingly_parallel
        A boolean indicating whether any implementation in the pipeline is to be
        run in an embarrassingly parallel manner.

    """

    def __init__(self, config: Config):
        self.config = config
        self.pipeline_graph = PipelineGraph(config)
        self.spark_is_required = self.pipeline_graph.spark_is_required
        self.any_embarrassingly_parallel = self.pipeline_graph.any_embarrassingly_parallel

        # TODO [MIC-4880]: refactor into validation object
        self._validate()

    @property
    def snakefile_path(self) -> Path:
        """The path to the dynamically-generated snakefile."""
        return self.config.results_dir / "Snakefile"

    def build_snakefile(self) -> Path:
        """Generates the Snakefile for this ``Pipeline``.

        This method dynamically builds the Snakefile by generating all necessary
        setup instructions (e.g. imports, configuration settings) as well as
        all rules for each :class:`~easylink.implementation.Implementation` in the
        pipeline and appending them to the Snakefile.

        Returns
        -------
            The path to the Snakefile.

        Notes
        -----
        We use the Snakemake term "rule" to refer to a singular component in a Snakefile
        (i.e. in a Snakemake pipeline) that defines input files, output files,
        and the command to run to create those output files. These rules are generated
        dynamically as strings and appended to the Snakefile.
        """
        if self.snakefile_path.is_file():
            logger.warning("Snakefile already exists, overwriting.")
            self.snakefile_path.unlink()
        self._write_imports()
        self._write_wildcard_constraints()
        self._write_spark_config()
        self._write_target_rules()
        self._write_spark_module()
        for node in self.pipeline_graph.implementation_nodes:
            self._write_implementation_rules(node)
        return self.snakefile_path

    ##################
    # Helper Methods #
    ##################

    def _validate(self) -> None:
        """Validates the pipeline.

        Raises
        ------
        SystemExit
            If any errors are found, they are batch-logged into a dictionary and
            the program exits with a non-zero code.
        """

        errors = {**self._validate_implementations()}

        if errors:
            exit_with_validation_error(errors)

    def _validate_implementations(self) -> dict:
        """Validates each individual :class:`~easylink.implementation.Implementation` instance.

        Returns
        -------
            A dictionary of ``Implementation`` validation errors.
        """
        errors = defaultdict(dict)
        for implementation in self.pipeline_graph.implementations:
            implementation_errors = implementation.validate()
            if implementation_errors:
                errors[IMPLEMENTATION_ERRORS_KEY][implementation.name] = implementation_errors
        return errors

    def _write_imports(self) -> None:
        if not self.any_embarrassingly_parallel:
            imports = "from easylink.utilities import validation_utils\n"
        else:
            imports = """import glob
import os

from snakemake.exceptions import IncompleteCheckpointException
from snakemake.io import checkpoint_target

from easylink.utilities import aggregator_utils, splitter_utils, validation_utils\n"""
        with open(self.snakefile_path, "a") as f:
            f.write(imports)

    def _write_wildcard_constraints(self) -> None:
        if self.any_embarrassingly_parallel:
            with open(self.snakefile_path, "a") as f:
                f.write(
                    """
wildcard_constraints:
    # never include '/' since those are reserved for filepaths
    chunk="[^/]+",\n"""
                )

    def _write_target_rules(self) -> None:
        """Writes the rule for the final output and its validation.

        The input files to the the target rule (i.e. the result node) are the final
        output themselves.
        """
        final_output, _ = self.pipeline_graph.get_io_filepaths("results")
        validator_file = str("input_validations/final_validator")
        # Snakemake resolves the DAG based on the first rule, so we put the target
        # before the validation
        target_rule = TargetRule(
            target_files=final_output,
            validation=validator_file,
            requires_spark=self.spark_is_required,
        )
        final_validation = InputValidationRule(
            name="results",
            input_slot_name="main_input",
            input=final_output,
            output=validator_file,
            validator=validate_input_file_dummy,
        )
        target_rule.write_to_snakefile(self.snakefile_path)
        final_validation.write_to_snakefile(self.snakefile_path)

    def _write_spark_config(self) -> None:
        """Writes configuration settings to the Snakefile.

        Notes
        -----
        This is currently only applicable for spark-dependent pipelines.
        """
        if self.spark_is_required:
            with open(self.snakefile_path, "a") as f:
                f.write(
                    f"\nscattergather:\n\tnum_workers={self.config.spark_resources['num_workers']},"
                )

    def _write_spark_module(self) -> None:
        """Inserts the ``easylink.utilities.spark.smk`` Snakemake module into the Snakefile."""
        if not self.spark_is_required:
            return
        slurm_resources = self.config.slurm_resources
        spark_resources = self.config.spark_resources
        module = f"""
module spark_cluster:
    snakefile: '{SPARK_SNAKEFILE}'
    config: config

use rule * from spark_cluster
use rule terminate_spark from spark_cluster with:
    input: rules.all.input.final_output"""
        if self.config.computing_environment == "slurm":
            module += f"""
use rule start_spark_master from spark_cluster with:
    resources:
        slurm_account={slurm_resources['slurm_account']},
        slurm_partition={slurm_resources['slurm_partition']},
        mem_mb={spark_resources['slurm_mem_mb']},
        runtime={spark_resources['runtime']},
        cpus_per_task={spark_resources['cpus_per_task']},
        slurm_extra="--output 'spark_logs/start_spark_master-slurm-%j.log'"
use rule start_spark_worker from spark_cluster with:
    resources:
        slurm_account={slurm_resources['slurm_account']},
        slurm_partition={slurm_resources['slurm_partition']},
        mem_mb={spark_resources['slurm_mem_mb']},
        runtime={spark_resources['runtime']},
        cpus_per_task={spark_resources['cpus_per_task']},
        slurm_extra="--output 'spark_logs/start_spark_worker-slurm-%j.log'"
    params:
        terminate_file_name=rules.terminate_spark.output,
        user=os.environ["USER"],
        cores={spark_resources['cpus_per_task']},
        memory={spark_resources['mem_mb']}"""

        with open(self.snakefile_path, "a") as f:
            f.write(module)

    def _write_implementation_rules(self, node_name: str) -> None:
        """Writes the rules for each :class:`~easylink.implementation.Implementation`.

        This method writes *all* rules required for a given ``Implementation``,
        e.g. splitters and aggregators (if necessary), validations, and the actual
        rule to run the container itself.

        Parameters
        ----------
        node_name
            The name of the ``Implementation`` to write the rule(s) for.
        """

        input_slots, output_slots = self.pipeline_graph.get_io_slot_attributes(node_name)
        validation_files, validation_rules = self._get_validations(node_name, input_slots)
        for validation_rule in validation_rules:
            validation_rule.write_to_snakefile(self.snakefile_path)

        _input_files, output_files = self.pipeline_graph.get_io_filepaths(node_name)
        is_embarrassingly_parallel = self.pipeline_graph.get_whether_embarrassingly_parallel(
            node_name
        )
        if is_embarrassingly_parallel:
            CheckpointRule(
                name=node_name,
                input_slots=input_slots,
                validations=validation_files,
                output=output_files,
            ).write_to_snakefile(self.snakefile_path)
            for name, attrs in output_slots.items():
                AggregationRule(
                    name=node_name,
                    input_slots=input_slots,
                    output_slot_name=name,
                    output_slot=attrs,
                ).write_to_snakefile(self.snakefile_path)

        implementation = self.pipeline_graph.nodes[node_name]["implementation"]
        diagnostics_dir = Path("diagnostics") / node_name
        diagnostics_dir.mkdir(parents=True, exist_ok=True)
        resources = (
            self.config.slurm_resources
            if self.config.computing_environment == "slurm"
            else None
        )
        ImplementedRule(
            name=node_name,
            step_name=" and ".join(implementation.metadata_steps),
            implementation_name=implementation.name,
            input_slots=input_slots,
            validations=validation_files,
            output=output_files,
            resources=resources,
            envvars=implementation.environment_variables,
            diagnostics_dir=str(diagnostics_dir),
            image_path=implementation.singularity_image_path,
            script_cmd=implementation.script_cmd,
            requires_spark=implementation.requires_spark,
            is_embarrassingly_parallel=is_embarrassingly_parallel,
        ).write_to_snakefile(self.snakefile_path)

    @staticmethod
    def _get_validations(
        node_name, input_slots
    ) -> tuple[list[str], list[InputValidationRule]]:
        """Gets the validation rule and its output filepath for each slot for a given node.

        Parameters
        ----------
        node_name
            The name of the ``Implementation`` to get validation data for.
        input_slots
            The input slot attributes for the given node.

        Returns
        -------
            A tuple of lists containing the validation output paths and rules.
        """
        validation_files = []
        validation_rules = []

        for input_slot_name, input_slot_attrs in input_slots.items():
            validation_file = f"input_validations/{node_name}/{input_slot_name}_validator"
            validation_files.append(validation_file)
            validation_rules.append(
                InputValidationRule(
                    name=node_name,
                    input_slot_name=input_slot_name,
                    input=input_slot_attrs["filepaths"],
                    output=validation_file,
                    validator=input_slot_attrs["validator"],
                )
            )
        return validation_files, validation_rules



# mypy: ignore-errors
import os
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from easylink.rule import (
    AggregationRule,
    CheckpointRule,
    ImplementedRule,
    InputValidationRule,
    Rule,
    TargetRule,
)

RULE_STRINGS = {
    "target_rule": "rule_strings/target_rule.txt",
    "implemented_rule_local": "rule_strings/implemented_rule_local.txt",
    "implemented_rule_slurm": "rule_strings/implemented_rule_slurm.txt",
    "validation_rule": "rule_strings/validation_rule.txt",
    "embarrassingly_parallel_rule": "rule_strings/embarrassingly_parallel_rule.txt",
    "checkpoint_rule": "rule_strings/checkpoint_rule.txt",
    "aggregation_rule": "rule_strings/aggregation_rule.txt",
}


def test_rule_interface():
    class TestRule(Rule):
        def build_rule(self) -> str:
            return "'Tis but a scratch!"

    rule = TestRule()
    with TemporaryDirectory() as snake_dir_path:
        snakefile_path = Path(snake_dir_path) / "Snakefile"
        rule.write_to_snakefile(snakefile_path)
        with open(snakefile_path, "r") as f:
            assert f.read() == "'Tis but a scratch!"


def test_target_rule_build_rule():
    rule = TargetRule(
        target_files=["int/foo", "int/bar"],
        validation="bar",
        requires_spark=False,
    )
    file_path = Path(os.path.dirname(__file__)) / RULE_STRINGS["target_rule"]
    _check_rule(rule, file_path)


@pytest.mark.parametrize("computing_environment", ["local", "slurm"])
def test_implemented_rule_build_rule(computing_environment):
    if computing_environment == "slurm":
        resources = {
            "slurm_partition": "'slurmpart'",
            "runtime": 1,
            "mem_mb": 5120,
            "cpus_per_task": 1337,
        }
    else:
        resources = None

    rule = ImplementedRule(
        name="foo_rule",
        step_name="foo_step",
        implementation_name="foo_imp",
        input_slots={
            "main": {
                "env_var": "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                "filepaths": ["foo"],
                "validator": _dummy_validator,
            },
            "secondary": {
                "env_var": "DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                "filepaths": ["bar"],
                "validator": _dummy_validator,
            },
        },
        validations=["bar"],
        output=["baz"],
        resources=resources,
        envvars={"eggs": "coconut"},
        diagnostics_dir="spam",
        image_path="Multipolarity.sif",
        script_cmd="echo hello world",
        requires_spark=False,
    )

    file_path = (
        Path(os.path.dirname(__file__))
        / RULE_STRINGS[f"implemented_rule_{computing_environment}"]
    )
    _check_rule(rule, file_path)


def test_embarrassingly_parallel_rule_build_rule():
    rule = ImplementedRule(
        name="foo_rule",
        step_name="foo_step",
        implementation_name="foo_imp",
        input_slots={
            "main": {
                "env_var": "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                "filepaths": ["foo"],
                "validator": _dummy_validator,
                "splitter": _dummy_splitter,
            },
            "secondary": {
                "env_var": "DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                "filepaths": ["bar"],
                "validator": _dummy_validator,
                "splitter": None,
            },
        },
        validations=["baz"],
        output=["some/path/to/quux"],
        resources=None,
        envvars={"eggs": "coconut"},
        diagnostics_dir="spam",
        image_path="Multipolarity.sif",
        script_cmd="echo hello world",
        requires_spark=False,
        is_embarrassingly_parallel=True,
    )

    file_path = Path(os.path.dirname(__file__)) / RULE_STRINGS["embarrassingly_parallel_rule"]
    _check_rule(rule, file_path)


def test_embarrassingly_parallel_rule_build_rule_multiple_outputs_raises():
    """Temporary test against raising if an embarrassingly parallel step has multiple outputs."""
    rule = ImplementedRule(
        name="foo_rule",
        step_name="foo_step",
        implementation_name="foo_imp",
        input_slots={
            "main": {
                "env_var": "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                "filepaths": ["foo"],
                "validator": _dummy_validator,
                "splitter": _dummy_splitter,
            },
        },
        validations=["baz"],
        output=["some/path/to/quux", "some/other/path/to/something"],
        resources=None,
        envvars={"eggs": "coconut"},
        diagnostics_dir="spam",
        image_path="Multipolarity.sif",
        script_cmd="echo hello world",
        requires_spark=False,
        is_embarrassingly_parallel=True,
    )

    with pytest.raises(
        NotImplementedError,
        match="Multiple output slots/files of EmbarrassinglyParallelSteps not yet supported",
    ):
        rule.build_rule()


def test_checkpoint_build_rule():
    rule = CheckpointRule(
        name="foo_rule",
        input_slots={
            "main": {
                "env_var": "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                "filepaths": ["foo"],
                "validator": _dummy_validator,
                "splitter": _dummy_splitter,
            },
            "secondary": {
                "env_var": "DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                "filepaths": ["bar"],
                "validator": _dummy_validator,
                "splitter": None,
            },
        },
        validations=["baz"],
        output=["some/path/to/quux"],
    )

    file_path = Path(os.path.dirname(__file__)) / RULE_STRINGS["checkpoint_rule"]
    _check_rule(rule, file_path)


def test_aggregation_build_rule():
    rule = AggregationRule(
        name="foo_rule",
        input_slots={
            "main": {
                "env_var": "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
                "filepaths": ["foo"],
                "validator": _dummy_validator,
                "splitter": _dummy_splitter,
            },
            "secondary": {
                "env_var": "DUMMY_CONTAINER_SECONDARY_INPUT_FILE_PATHS",
                "filepaths": ["bar"],
                "validator": _dummy_validator,
                "splitter": None,
            },
        },
        output_slot_name="main_output",
        output_slot={
            "filepaths": ["some/path/to/quux"],
            "aggregator": _dummy_aggregator,
        },
    )

    file_path = Path(os.path.dirname(__file__)) / RULE_STRINGS["aggregation_rule"]
    _check_rule(rule, file_path)


def test_validation_rule_build_rule():
    rule = InputValidationRule(
        name="foo",
        input_slot_name="foo_input",
        input=["foo", "bar"],
        output="baz",
        validator=_dummy_validator,
    )
    file_path = Path(os.path.dirname(__file__)) / RULE_STRINGS["validation_rule"]
    _check_rule(rule, file_path)


####################
# Helper functions #
####################


def _check_rule(rule: Rule, expected_rule_path: str):
    """Compares the ``Rule's`` built rule to the expected rule in the file."""
    with open(expected_rule_path) as expected_file:
        expected = expected_file.read()
    rulestring = rule.build_rule()
    rulestring_lines = rulestring.split("\n")
    expected_lines = expected.split("\n")
    assert len(rulestring_lines) == len(expected_lines)
    for i, expected_line in enumerate(expected_lines):
        assert rulestring_lines[i].strip() == expected_line.strip()


def _dummy_validator():
    """Dummy validator function"""
    pass


def _dummy_splitter():
    """Dummy splitter function"""
    pass


def _dummy_aggregator():
    """Dummy aggregator function"""
    pass



import glob
import logging
import os

import yaml
from pyspark.sql import SparkSession
from pyspark.sql.functions import col, lit
from pyspark.sql.functions import max as spark_max

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(message)s",
    handlers=[logging.StreamHandler()],
)

# https://stackoverflow.com/a/34248975/
pyspark_log = logging.getLogger("pyspark")
pyspark_log.setLevel(logging.WARNING)

spark = SparkSession.builder.master(
    os.getenv("DUMMY_CONTAINER_SPARK_MASTER_URL")
).getOrCreate()


def load_file(file_path, file_format=None):
    if file_format is None:
        file_format = file_path.split(".")[-1]
    if file_format == "parquet":
        return spark.read.parquet(file_path)
    if file_format == "csv":
        return spark.read.csv(file_path, header=True, inferSchema=True)
    raise ValueError()


diagnostics = {}

INPUT_ENV_VARS = os.getenv(
    "INPUT_ENV_VARS",
    "DUMMY_CONTAINER_MAIN_INPUT_FILE_PATHS",
).split(",")

df = None

for env_var in INPUT_ENV_VARS:
    if env_var not in os.environ:
        logging.error(f"Missing required environment variable {env_var}")
        raise ValueError()

    logging.info(f"Loading files for {env_var}")
    file_paths = os.environ[env_var].split(",")
    diagnostics[f"num_files_{env_var.lower()}"] = len(file_paths)

    if df is None:
        df = load_file(file_paths[0])
        file_paths = file_paths[1:]

    for path in file_paths:
        df = df.unionByName(load_file(path), allowMissingColumns=True).fillna(0)

extra_implementation_specific_input_glob = glob.glob(
    "/extra_implementation_specific_input_data/input*"
)
extra_implementation_specific_input_file_path = os.getenv(
    "DUMMY_CONTAINER_EXTRA_IMPLEMENTATION_SPECIFIC_INPUT_FILE_PATH",
    (
        extra_implementation_specific_input_glob[0]
        if len(extra_implementation_specific_input_glob) > 0
        else None
    ),
)
diagnostics["extra_implementation_specific_input"] = (
    extra_implementation_specific_input_file_path is not None
)
if extra_implementation_specific_input_file_path is not None:
    logging.info("Loading extra, implementation-specific input")
    df = df.unionByName(
        load_file(extra_implementation_specific_input_file_path), allowMissingColumns=True
    ).fillna(0)

logging.info(f"Total input length is {df.count()}")

broken = os.getenv("DUMMY_CONTAINER_BROKEN", "false").lower() in ("true", "yes", "1")
diagnostics["broken"] = broken
if broken:
    df = (
        df.withColumnRenamed("foo", "wrong")
        .withColumnRenamed("bar", "column")
        .withColumnRenamed("counter", "names")
    )
else:
    increment = int(os.getenv("DUMMY_CONTAINER_INCREMENT", "1"))
    diagnostics["increment"] = increment
    logging.info(f"Increment is {increment}")
    df = df.withColumn("counter", df.counter + increment)

    added_columns_existing = [c for c in df.columns if "added_column_" in c]
    diagnostics["added_columns_existing"] = added_columns_existing
    added_columns_existing_ints = [int(c.split("_")[-1]) for c in added_columns_existing]
    max_added_column = max(added_columns_existing_ints, default=0) + increment
    diagnostics["max_added_column"] = max_added_column
    min_added_column = max(max_added_column - 4, 0)
    added_columns_desired = range(min_added_column, max_added_column + 1)
    added_columns_desired_names = [f"added_column_{i}" for i in added_columns_desired]
    diagnostics["added_columns_desired_names"] = added_columns_desired_names
    diagnostics["new_columns"] = []
    for column_index, column_name in zip(added_columns_desired, added_columns_desired_names):
        if column_name not in df.columns:
            diagnostics["new_columns"].append(column_name)
            df = df.withColumn(column_name, lit(column_index))

    columns_to_drop = [
        c for c in df.columns if "added_column_" in c and c not in added_columns_desired_names
    ]
    diagnostics["columns_to_drop"] = columns_to_drop
    df = df.drop(*columns_to_drop)

output_file_format = os.getenv("DUMMY_CONTAINER_OUTPUT_FILE_FORMAT", "parquet")
output_file_paths = os.getenv(
    "DUMMY_CONTAINER_OUTPUT_PATHS", f"/results/result.{output_file_format}"
).split(",")

diagnostics["num_output_files"] = len(output_file_paths)
diagnostics["output_file_paths"] = output_file_paths

for output_file_path in output_file_paths:
    logging.info(f"Writing output to {output_file_path} in {output_file_format} format")
    # NOTE: Go back to pandas in order to save a single file
    if output_file_format == "parquet":
        df.toPandas().to_parquet(output_file_path)
    elif output_file_format == "csv":
        df.toPandas().to_csv(output_file_path, index=False)
    else:
        raise ValueError()

diagnostics_dir = os.getenv("DUMMY_CONTAINER_DIAGNOSTICS_DIRECTORY", "/diagnostics")
try:
    with open(f"{diagnostics_dir}/diagnostics.yaml", "w") as f:
        yaml.dump(diagnostics, f, default_flow_style=False)
except (PermissionError, OSError):
    pass



# mypy: ignore-errors
import os
import shutil
from datetime import datetime
from pathlib import Path

import yaml


def modify_umask(func):
    """Decorator to wrap umask modification before making directories"""

    def wrapper(*args, **kwargs):
        old_umask = os.umask(0o002)
        try:
            return func(*args, **kwargs)
        finally:
            os.umask(old_umask)

    return wrapper


@modify_umask
def create_results_directory(results_dir: Path) -> None:
    results_dir.mkdir(parents=True, exist_ok=True)


@modify_umask
def create_results_intermediates(results_dir: Path) -> None:
    (results_dir / "intermediate").mkdir(exist_ok=True)
    (results_dir / "diagnostics").mkdir(exist_ok=True)


def copy_configuration_files_to_results_directory(
    pipeline_specification: Path,
    input_data: Path,
    computing_environment: Path | None,
    results_dir: Path,
) -> None:
    shutil.copy(pipeline_specification, results_dir)
    shutil.copy(input_data, results_dir)
    if computing_environment:
        shutil.copy(computing_environment, results_dir)


def get_results_directory(output_dir: str | None, no_timestamp: bool) -> Path:
    results_dir = Path("results" if output_dir is None else output_dir).resolve()
    if not no_timestamp:
        launch_time = _get_timestamp()
        results_dir = results_dir / launch_time
    return results_dir


def _get_timestamp():
    return datetime.now().strftime("%Y_%m_%d_%H_%M_%S")


def load_yaml(filepath: str | Path) -> dict:
    with open(filepath, "r") as file:
        data = yaml.safe_load(file)
    return data



# mypy: ignore-errors
from pathlib import Path
from typing import Any

import pytest
from layered_config_tree import LayeredConfigTree

from easylink.configuration import (
    DEFAULT_ENVIRONMENT,
    SPARK_DEFAULTS,
    Config,
    _load_computing_environment,
    _load_input_data_paths,
)
from easylink.pipeline_schema import PIPELINE_SCHEMAS


def test__get_schema(default_config: Config) -> None:
    """Test default config gets "development schema", without errors"""
    assert default_config.schema == PIPELINE_SCHEMAS[0]


def test_load_params_from_specification(
    test_dir: str, default_config_params: dict[str, dict[str, Any]]
) -> None:
    assert default_config_params == {
        "pipeline": {
            "steps": {
                "step_1": {"implementation": {"name": "step_1_python_pandas"}},
                "step_2": {"implementation": {"name": "step_2_python_pandas"}},
                "step_3": {"implementation": {"name": "step_3_python_pandas"}},
                "choice_section": {
                    "type": "simple",
                    "step_4": {"implementation": {"name": "step_4_python_pandas"}},
                },
            }
        },
        "input_data": {
            "file1": Path(f"{test_dir}/input_data1/file1.csv"),
            "file2": Path(f"{test_dir}/input_data2/file2.csv"),
        },
        "environment": {
            "computing_environment": "local",
            "container_engine": "undefined",
        },
        "results_dir": Path(f"{test_dir}/results_dir"),
    }


def test__load_input_data_paths(test_dir: str) -> None:
    paths = _load_input_data_paths(f"{test_dir}/input_data.yaml")
    assert paths == {
        "file1": Path(f"{test_dir}/input_data1/file1.csv"),
        "file2": Path(f"{test_dir}/input_data2/file2.csv"),
    }


@pytest.mark.parametrize(
    "environment_file, expected",
    [
        # good
        (
            "environment.yaml",
            {
                k: v
                for k, v in DEFAULT_ENVIRONMENT["environment"].items()
                if k in ["computing_environment", "container_engine"]
            },
        ),
        (None, {}),
    ],
)
def test__load_computing_environment(test_dir, environment_file, expected):
    filepath = Path(f"{test_dir}/{environment_file}") if environment_file else None
    env = _load_computing_environment(filepath)
    assert env == expected


def test_load_missing_computing_environment_fails():
    with pytest.raises(
        FileNotFoundError,
        match="Computing environment is expected to be a path to an existing specification file. .*",
    ):
        _load_computing_environment(Path("some/bogus/path.yaml"))


def test_input_data_configuration_requires_key_value_pairs(test_dir):
    with pytest.raises(
        TypeError, match="Input data should be submitted like 'key': path/to/file."
    ):
        _load_input_data_paths(f"{test_dir}/input_data_list.yaml")


@pytest.mark.parametrize(
    "computing_environment",
    [
        "bad/path/to/environment.yaml",
        Path("another/bad/path"),
    ],
)
def test_environment_configuration_not_found(computing_environment):
    with pytest.raises(FileNotFoundError):
        _load_computing_environment(computing_environment)


@pytest.mark.parametrize(
    "key, value",
    [
        ("computing_environment", None),
        ("computing_environment", "local"),
        ("computing_environment", "slurm"),
        ("container_engine", None),
        ("container_engine", "docker"),
        ("container_engine", "singularity"),
        ("container_engine", "undefined"),
    ],
)
def test_required_attributes(mocker, default_config_params, key, value):
    mocker.patch("easylink.configuration.Config._validate_environment")
    config_params = default_config_params
    if value:
        config_params["environment"][key] = value
    env_dict = {key: value} if value else {}
    retrieved = Config(config_params).environment[key]
    expected = DEFAULT_ENVIRONMENT["environment"].copy()
    expected.update(env_dict)
    assert retrieved == expected[key]


@pytest.mark.parametrize(
    "resource_request",
    [
        # missing
        None,
        # partially defined
        {"memory": 100},
        # fully defined
        {"memory": 100, "cpus": 200, "time_limit": 300},
    ],
)
def test_implementation_resource_requests(default_config_params, resource_request):
    key = "implementation_resources"
    config_params = default_config_params
    if resource_request:
        config_params["environment"][key] = resource_request
    config = Config(config_params)
    env_dict = {key: resource_request.copy()} if resource_request else {}
    retrieved = config.environment[key].to_dict()
    expected = DEFAULT_ENVIRONMENT["environment"][key].copy()
    if resource_request:
        expected.update(env_dict[key])
    assert retrieved == expected


@pytest.mark.parametrize(
    "spark_request",
    [
        # missing
        None,
        # partially defined (missing entire workers)
        {"keep_alive": "idk"},
        # partially defined (missing keep_alive and num_workers)
        {
            "workers": {
                "cpus_per_node": 200,
                "mem_per_node": 300,
                "time_limit": 400,
            }
        },
        # fully defined
        {
            "workers": {
                "num_workers": 100,
                "cpus_per_node": 200,
                "mem_per_node": 300,
                "time_limit": 400,
            },
            "keep_alive": "idk",
        },
    ],
)
@pytest.mark.parametrize("requires_spark", [True, False])
def test_spark_requests(default_config_params, spark_request, requires_spark):
    key = "spark"
    config_params = default_config_params
    if requires_spark:
        # Change step 1's implementation to python_pyspark
        config_params["pipeline"]["steps"]["step_1"]["implementation"][
            "name"
        ] = "step_1_python_pyspark"

    if spark_request:
        config_params["environment"][key] = spark_request
    retrieved = Config(config_params).environment[key].to_dict()
    expected_env_dict = {key: spark_request.copy()} if spark_request else {}
    expected = LayeredConfigTree(SPARK_DEFAULTS, layers=["initial_data", "user"])
    if spark_request:
        expected.update(expected_env_dict[key], layer="user")
    expected = expected.to_dict()
    assert retrieved == expected


@pytest.mark.parametrize("is_default", [True, False])
def test_combined_implementations(default_config_params, is_default):
    combined_dict = {"foo": "bar"}
    config_params = default_config_params
    if not is_default:
        config_params["pipeline"]["combined_implementations"] = combined_dict
    combined_implementations = Config(config_params).pipeline.combined_implementations
    expected = {} if is_default else combined_dict
    assert combined_implementations.to_dict() == expected
